# Unit test create_train_X_y ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect


@pytest.mark.parametrize("step", [0, 4], ids=lambda step: f'step: {step}')
def test_filter_train_X_y_for_step_ValueError_when_step_not_in_steps(step):
    """
    Test ValueError is raised when step not in steps.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=3
    )
    X_train, y_train = forecaster.create_train_X_y(y)

    err_msg = re.escape(
        f"Invalid value `step`. For this forecaster, minimum value is 1 "
        f"and the maximum step is {forecaster.max_step}."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.filter_train_X_y_for_step(step=step, X_train=X_train, y_train=y_train)


def test_filter_train_X_y_for_step_output_when_lags_3_steps_2_exog_is_None_for_step_1():
    """
    Test output of filter_train_X_y_for_step when estimator is LinearRegression, 
    lags is 3 and steps is 2 for step 1.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=2
    )
    X_train, y_train = forecaster.create_train_X_y(y=y)
    results = forecaster.filter_train_X_y_for_step(step=1, X_train=X_train, y_train=y_train)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0.],
                             [3., 2., 1.],
                             [4., 3., 2.],
                             [5., 4., 3.],
                             [6., 5., 4.],
                             [7., 6., 5.]], dtype=float),
            index   = pd.RangeIndex(start=3, stop=9, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3']
        ),
        pd.Series(
            data  = np.array([3., 4., 5., 6., 7., 8.], dtype=float),
            index = pd.RangeIndex(start=3, stop=9, step=1),
            name  = 'y_step_1'
        )
    )  
 
    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])


def test_filter_train_X_y_for_step_output_when_lags_3_steps_2_and_exog_for_step_2():
    """
    Test output of filter_train_X_y_for_step when estimator is LinearRegression, 
    lags is 3 and steps is 2 with exog for step 2.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=float)

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=2
    )
    X_train, y_train = forecaster.create_train_X_y(y=y, exog=exog)
    results = forecaster.filter_train_X_y_for_step(step=2, X_train=X_train, y_train=y_train)

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0., 104.],
                             [3., 2., 1., 105.],
                             [4., 3., 2., 106.],
                             [5., 4., 3., 107.],
                             [6., 5., 4., 108.],
                             [7., 6., 5., 109.]], dtype=float),
            index   = pd.RangeIndex(start=4, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'exog_step_2']
        ),
        pd.Series(
            data  = np.array([4., 5., 6., 7., 8., 9.]),
            index = pd.RangeIndex(start=4, stop=10, step=1),
            name  = 'y_step_2', 
            dtype = float
        )
    )  
 
    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])


def test_filter_train_X_y_for_step_output_when_lags_3_steps_2_and_exog_for_step_2_remove_suffix():
    """
    Test output of filter_train_X_y_for_step when estimator is LinearRegression, 
    lags is 3 and steps is 2 with exog for step 2 with remove_suffix=True.
    """
    y = pd.Series(np.arange(10), name='y', dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=float)

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=2
    )
    X_train, y_train = forecaster.create_train_X_y(y=y, exog=exog)
    results = forecaster.filter_train_X_y_for_step(
                  step          = 2, 
                  X_train       = X_train, 
                  y_train       = y_train,
                  remove_suffix = True
              )

    expected = (
        pd.DataFrame(
            data = np.array([[2., 1., 0., 104.],
                             [3., 2., 1., 105.],
                             [4., 3., 2., 106.],
                             [5., 4., 3., 107.],
                             [6., 5., 4., 108.],
                             [7., 6., 5., 109.]], dtype=float),
            index   = pd.RangeIndex(start=4, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'exog']
        ),
        pd.Series(
            data  = np.array([4., 5., 6., 7., 8., 9.]),
            index = pd.RangeIndex(start=4, stop=10, step=1),
            name  = 'y', 
            dtype = float
        )
    )
 
    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])


def test_filter_train_X_y_for_step_output_when_window_features_and_exog_steps_1():
    """
    Test the output of filter_train_X_y_for_step when using window_features and exog 
    with datetime index and steps=1.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=1, lags=5, window_features=rolling
    )
    X_train, y_train = forecaster.create_train_X_y(y=y_datetime, exog=exog_datetime)
    results = forecaster.filter_train_X_y_for_step(
                  step          = 1, 
                  X_train       = X_train, 
                  y_train       = y_train,
                  remove_suffix = False
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 106.],
                             [6., 5., 4., 3., 2., 4., 4., 21., 107.],
                             [7., 6., 5., 4., 3., 5., 5., 27., 108.],
                             [8., 7., 6., 5., 4., 6., 6., 33., 109.],
                             [9., 8., 7., 6., 5., 7., 7., 39., 110.],
                             [10., 9., 8., 7., 6., 8., 8., 45., 111.],
                             [11., 10., 9., 8., 7., 9., 9., 51., 112.],
                             [12., 11., 10., 9., 8., 10., 10., 57., 113.],
                             [13., 12., 11., 10., 9., 11., 11., 63., 114.]]),
            index   = pd.date_range('2000-01-07', periods=9, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 'exog_step_1']
        ),
        pd.Series(
            data  = np.array([6, 7, 8, 9, 10, 11, 12, 13, 14]),
            index = pd.date_range('2000-01-07', periods=9, freq='D'),
            name  = 'y_step_1',
            dtype = float
        )
    )
 
    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])


def test_filter_train_X_y_for_step_output_when_window_features_and_exog_steps_2():
    """
    Test the output of filter_train_X_y_for_step when using window_features and 
    exog with datetime index and steps=2.
    """
    y_datetime = pd.Series(
        np.arange(15), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='y', dtype=float
    )
    exog_datetime = pd.Series(
        np.arange(100, 115), index=pd.date_range('2000-01-01', periods=15, freq='D'),
        name='exog', dtype=float
    )
    rolling = RollingFeatures(
        stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
    )

    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=2, lags=5, window_features=rolling
    )
    X_train, y_train = forecaster.create_train_X_y(y=y_datetime, exog=exog_datetime)
    results = forecaster.filter_train_X_y_for_step(
                  step          = 2, 
                  X_train       = X_train, 
                  y_train       = y_train,
                  remove_suffix = True
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[5., 4., 3., 2., 1., 3., 3., 15., 107.],
                             [6., 5., 4., 3., 2., 4., 4., 21., 108.],
                             [7., 6., 5., 4., 3., 5., 5., 27., 109.],
                             [8., 7., 6., 5., 4., 6., 6., 33., 110.],
                             [9., 8., 7., 6., 5., 7., 7., 39., 111.],
                             [10., 9., 8., 7., 6., 8., 8., 45., 112.],
                             [11., 10., 9., 8., 7., 9., 9., 51., 113.],
                             [12., 11., 10., 9., 8., 10., 10., 57., 114.]]),
            index   = pd.date_range('2000-01-08', periods=8, freq='D'),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       'roll_mean_5', 'roll_median_5', 'roll_sum_6', 
                       'exog']
        ),
        pd.Series(
            data  = np.array([7, 8, 9, 10, 11, 12, 13, 14]),
            index = pd.date_range('2000-01-08', periods=8, freq='D'),
            name  = 'y',
            dtype = float
        )
    )
 
    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])



# Unit test backtesting_gif_creator
# ==============================================================================
import os
import re
import pytest
import numpy as np
import pandas as pd
import matplotlib
from skforecast.model_selection import TimeSeriesFold
from ... import backtesting_gif_creator


def test_backtesting_gif_creator_raise_error_invalid_arguments():
    """
    Test that backtesting_gif_creator raises an error when invalid 
    arguments are passed.
    """

    # Don't use interactive backends during testing
    matplotlib.use("Agg")

    data = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    cv = TimeSeriesFold(
             initial_train_size=5,
             steps=2,
         )

    wrong_data = np.arange(10)
    err_msg = re.escape(
        f"`data` must be a pandas Series or DataFrame. Got {type(wrong_data)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=wrong_data, cv=cv)

    wrong_data = pd.DataFrame(np.arange(10))
    wrong_data.index = pd.Index(np.arange(10))
    err_msg = re.escape(
        f"`data` must have a pandas RangeIndex or DatetimeIndex. Got {type(wrong_data.index)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=wrong_data, cv=cv)

    wrong_cv = object()
    err_msg = re.escape(
        f"`cv` must be a 'TimeSeriesFold' object. Got '{type(wrong_cv).__name__}'."
    )
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=data, cv=wrong_cv)

    fps = -1
    err_msg = re.escape(f"`fps` must be a positive integer. Got {fps}.")
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=data, cv=cv, fps=fps)

    dpi = -1
    err_msg = re.escape(f"`dpi` must be a positive integer. Got {dpi}.")
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=data, cv=cv, dpi=dpi)

    series_to_plot = "x"
    err_msg = re.escape(
        f"`series_to_plot` must be a list of column names. Got {type(series_to_plot)}."
    )
    with pytest.raises(TypeError, match=err_msg):
        backtesting_gif_creator(data=data, cv=cv, series_to_plot=series_to_plot)

    series_to_plot = ["x"]
    err_msg = re.escape("Columns not found in `data`: ['x']")
    with pytest.raises(ValueError, match=err_msg):
        backtesting_gif_creator(data=data, cv=cv, series_to_plot=series_to_plot)

    wrong_colors = {
        "train": "#329239",
        "test": "#0d579b",
    }
    err_msg = re.escape(
        f"`colors` must contain the following keys: "
        f"{['train', 'last_window', 'gap', 'test', 'v_lines']}"
    )
    with pytest.raises(ValueError, match=err_msg):
        backtesting_gif_creator(data=data, cv=cv, colors=wrong_colors)


@pytest.mark.parametrize(
    "cv_kwargs", 
    [{'steps': 10, 'initial_train_size': 70},
     {'steps': 10, 'initial_train_size': 70, 'refit': True},
     {'steps': 10, 'initial_train_size': 70, 'refit': True, 'fixed_train_size': False},
     {'steps': 10, 'initial_train_size': 70, 'gap': 5},
     {'steps': 8, 'initial_train_size': 70, 'gap': 5, 'allow_incomplete_fold': False, 'refit': 2},
     {'steps': 8, 'initial_train_size': 70, 'skip_folds': 2, 'allow_incomplete_fold': False},
     {'steps': 8, 'initial_train_size': 70, 'fold_stride': 5}
     ], 
    ids=lambda cv_kwargs: f'cv_kwargs: {cv_kwargs}'
)
def test_backtesting_gif_creator_output(cv_kwargs):
    """
    Check that backtesting_gif_creator returns the expected output.
    """
    data = pd.DataFrame({
        'y': np.arange(0, 100),
        'exog': np.arange(100, 200)
    })
    cv = TimeSeriesFold(**cv_kwargs, verbose=False)
    cv_window_size = TimeSeriesFold(**cv_kwargs, window_size=5, verbose=False)

    backtesting_gif_creator(data=data['y'], cv=cv, filename="backtesting.gif", dpi=50)
    backtesting_gif_creator(data=data, cv=cv_window_size, filename="backtesting.gif", dpi=50)
    os.remove("backtesting.gif")



# Unit test transform_series
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from skforecast.utils import transform_series


def test_transform_series_TypeError_when_series_is_not_pandas_series():
    """
    Test TypeError is raised when series is not a pandas series.
    """
    series = np.arange(10)

    err_msg = re.escape(f"`series` argument must be a pandas Series. Got {type(series)}")
    with pytest.raises(TypeError, match = err_msg):
        transform_series(
            series            = series,
            transformer       = None,
            fit               = True,
            inverse_transform = False
        )


def test_transform_series_when_transformer_is_None():
    """
    Test the output of transform_series when transformer is None.
    """
    input_series = pd.Series([1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49])
    expected = input_series
    transformer = None
    results = transform_series(
                  series = input_series,
                  transformer = transformer,
                  fit = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_series_equal(results, expected)


def test_transform_series_when_transformer_is_StandardScaler():
    """
    Test the output of transform_series when transformer is StandardScaler.
    """
    input_series = pd.Series([1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49])
    expected = pd.Series([0.67596768, -0.47871021, -0.19805933,  1.67027365, -0.0537246,
                         -0.70323091, -1.39283021,  0.75615365,  1.17312067, -1.44896038],
                         name = 'no_name')
    transformer = StandardScaler()
    results = transform_series(
                  series = input_series,
                  transformer = transformer,
                  fit = True,
                  inverse_transform = False
              )
    
    pd.testing.assert_series_equal(results, expected)


def test_transform_series_when_transformer_is_StandardScaler_and_inverse_transform_is_True():
    """
    Test the output of transform_series when transformer is StandardScaler and
    inverse_transform is True.
    """
    input_1 = pd.Series([1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49],
                      name = 'y')
    transformer = StandardScaler()
    transformer.fit(input_1.to_numpy().reshape(-1, 1))
    input_2 = transformer.transform(input_1.to_numpy().reshape(-1, 1))
    input_2 = pd.Series(input_2.ravel(), name= 'y')
    expected = input_1
    results = transform_series(
                  series = input_2,
                  transformer = transformer,
                  fit = False,
                  inverse_transform = True
              )
    
    pd.testing.assert_series_equal(results, expected)


def test_transform_series_when_transformer_is_OneHotEncoder():
    """
    Test the output of transform_series when transformer is OneHotEncoder.
    """
    input_series = pd.Series(["A"] * 5 + ["B"] * 5, name='exog')
    transformer = OneHotEncoder(sparse_output=False)
    transformer.fit(input_series.to_frame())
    results = transform_series(
                  series = input_series,
                  transformer = transformer,
                  fit = False,
                  inverse_transform = False
              )

    expected = pd.DataFrame(
                data = np.array(
                            [[1., 0.],
                            [1., 0.],
                            [1., 0.],
                            [1., 0.],
                            [1., 0.],
                            [0., 1.],
                            [0., 1.],
                            [0., 1.],
                            [0., 1.],
                            [0., 1.]]
                       ),
                columns = ['exog_A', 'exog_B']
    )
    
    pd.testing.assert_frame_equal(results, expected)


def test_transform_series_when_applied_to_serie_with_different_name_than_the_one_used_to_fit():
    """
    Test the transform series works when the serie to transform has different name
    than the serie used for training.
    """
    training_series = pd.Series([1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49],
                                name = 'y')
    input_series = pd.Series([1.16, -0.28, 0.07, 2.4, 0.25, -0.56, -1.42, 1.26, 1.78, -1.49],
                             name = 'pred')
    expected = pd.Series([0.67596768, -0.47871021, -0.19805933,  1.67027365, -0.0537246,
                         -0.70323091, -1.39283021,  0.75615365,  1.17312067, -1.44896038],
                         name = 'pred')
    transformer = StandardScaler()
    transformer.fit(training_series.to_frame())
    results = transform_series(
                  series = input_series,
                  transformer = transformer,
                  fit = False,
                  inverse_transform = False
              )
    
    pd.testing.assert_series_equal(results, expected)


# Unit test reshape_exog_long_to_dict
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
import platform
from ...preprocessing import reshape_exog_long_to_dict
from ....exceptions import MissingValuesWarning

# Fixtures
from .fixtures_preprocessing import exog_A, exog_B, exog_C, n_exog_A, n_exog_B, n_exog_C
from .fixtures_preprocessing import exog_long
exog_long_multiindex = exog_long.set_index(["series_id", "datetime"])


def test_TypeError_when_data_is_not_dataframe():
    """
    Raise TypeError if data is not a pandas DataFrame.
    """
    err_msg = "`data` must be a pandas DataFrame."
    with pytest.raises(TypeError, match=err_msg):
        reshape_exog_long_to_dict(
            data="not_a_dataframe",
            series_id="series_id",
            index="datetime",
            freq="D",
            drop_all_nan_cols=True,
            consolidate_dtypes=True,
        )


def test_ValueError_reshape_exog_long_to_dict_when_arguments_series_id_index_not_provided():
    """
    Check that ValueError is raised when the input dataframe does not have MultiIndex and the
    arguments `series_id`, `index` and `values` are not provided
    """

    err_msg = (
        "Arguments `series_id`, and `index` must be "
        "specified when the input DataFrame does not have a MultiIndex. "
        "Please provide a value for each of these arguments."
    )
    with pytest.raises(ValueError, match=err_msg):
        reshape_exog_long_to_dict(data=exog_long, freq="D")


def test_ValueError_when_series_id_not_in_data():
    """
    Raise ValueError if series_id is not in data.
    """
    series_id = "series_id_not_in_data"
    err_msg = f"Column '{series_id}' not found in `data`."
    with pytest.raises(ValueError, match=err_msg):
        reshape_exog_long_to_dict(
            data=exog_long,
            series_id=series_id,
            index="datetime",
            freq="D",
            drop_all_nan_cols=True,
            consolidate_dtypes=True,
        )


def test_ValueError_when_index_not_in_data():
    """
    Raise ValueError if index is not in data.
    """
    index = "series_id_not_in_data"
    err_msg = f"Column '{index}' not found in `data`."
    with pytest.raises(ValueError, match=err_msg):
        reshape_exog_long_to_dict(
            data=exog_long,
            series_id="series_id",
            index=index,
            freq="D",
            drop_all_nan_cols=True,
            consolidate_dtypes=True,
        )


def test_check_output_reshape_exog_long_to_dict_dropna_False():
    """
    Check output of reshape_exog_long_to_dict with dropna=False.
    """
    expected = {
        'A': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_A),
                'exog_2': np.nan,
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_A, freq="D")
        ),
        'B': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_B),
                'exog_2': 'b',
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_B, freq="D")
        ),
        'C': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_C),
                'exog_2': np.nan,
                'exog_3': 1.0
            },
            index=pd.date_range("2020-01-01", periods=n_exog_C, freq="D")
        )
    }

    for k in expected.keys():
        expected[k]['exog_1'] = expected[k]['exog_1'].astype(int)
        expected[k]['exog_2'] = expected[k]['exog_2'].astype(object)
        expected[k]['exog_3'] = expected[k]['exog_3'].astype(float)

    results = reshape_exog_long_to_dict(
        data=exog_long,
        series_id="series_id",
        index="datetime",
        freq="D",
        drop_all_nan_cols=False,
        consolidate_dtypes=True,
    )

    for k in expected.keys():
        pd.testing.assert_frame_equal(results[k], expected[k])


def test_check_output_reshape_exog_long_to_dict_dropna_False_when_multiindex():
    """
    Check output of reshape_exog_long_to_dict with dropna=False when input data is a MultiIndex DataFrame.
    """
    expected = {
        'A': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_A),
                'exog_2': np.nan,
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_A, freq="D")
        ),
        'B': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_B),
                'exog_2': 'b',
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_B, freq="D")
        ),
        'C': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_C),
                'exog_2': np.nan,
                'exog_3': 1.0
            },
            index=pd.date_range("2020-01-01", periods=n_exog_C, freq="D")
        )
    }

    for k in expected.keys():
        expected[k]['exog_1'] = expected[k]['exog_1'].astype(int)
        expected[k]['exog_2'] = expected[k]['exog_2'].astype(object)
        expected[k]['exog_3'] = expected[k]['exog_3'].astype(float)

    results = reshape_exog_long_to_dict(
        data=exog_long_multiindex,
        freq="D",
        drop_all_nan_cols=False,
        consolidate_dtypes=True,
    )

    for k in expected.keys():
        pd.testing.assert_frame_equal(results[k], expected[k])


def test_check_output_reshape_exog_long_to_dict_dropna_True():
    """
    Check output of reshape_exog_long_to_dict with dropna=True.
    """
    expected = {
        "A": exog_A.set_index("datetime").asfreq("D").drop(columns="series_id"),
        "B": exog_B.set_index("datetime").asfreq("D").drop(columns="series_id"),
        "C": exog_C.set_index("datetime").asfreq("D").drop(columns="series_id"),
    }

    for k in expected.keys():
        expected[k].index.name = None

    results = reshape_exog_long_to_dict(
        data=exog_long,
        series_id="series_id",
        index="datetime",
        freq="D",
        drop_all_nan_cols=True,
        consolidate_dtypes=True,
    )

    for k in expected.keys():
        pd.testing.assert_frame_equal(results[k], expected[k], check_dtype=False)


def test_check_output_reshape_exog_long_to_dict_dropna_True_when_multiindex():
    """
    Check output of reshape_exog_long_to_dict with dropna=True when input data is a MultiIndex DataFrame.
    """
    expected = {
        "A": exog_A.set_index("datetime").asfreq("D").drop(columns="series_id"),
        "B": exog_B.set_index("datetime").asfreq("D").drop(columns="series_id"),
        "C": exog_C.set_index("datetime").asfreq("D").drop(columns="series_id"),
    }

    for k in expected.keys():
        expected[k].index.name = None

    results = reshape_exog_long_to_dict(
        data=exog_long_multiindex,
        freq="D",
        drop_all_nan_cols=True,
        consolidate_dtypes=True,
    )

    for k in expected.keys():
        pd.testing.assert_frame_equal(results[k], expected[k], check_dtype=False)


def test_MissingValuesWarning_when_exog_are_incomplete_and_dropna_False():
    """
    Raise MissingValuesWarning if exogenous variables are incomplete and NaN values are introduced
    after setting the index frequency.
    """
    data = exog_long.copy().reset_index(drop=True)
    data = data.loc[[0, 1] + list(range(3, 30))]

    expected = {
        'A': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_A),
                'exog_2': np.nan,
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_A, freq="D")
        ),
        'B': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_B),
                'exog_2': 'b',
                'exog_3': np.nan
            },
            index=pd.date_range("2020-01-01", periods=n_exog_B, freq="D")
        ),
        'C': pd.DataFrame(
            {
                'exog_1': np.arange(n_exog_C),
                'exog_2': np.nan,
                'exog_3': 1.0
            },
            index=pd.date_range("2020-01-01", periods=n_exog_C, freq="D")
        )
    }
    expected['A'].loc['2020-01-03', 'exog_1'] = np.nan

    for k in expected.keys():
        expected[k]['exog_1'] = expected[k]['exog_1'].astype(float)
        expected[k]['exog_2'] = expected[k]['exog_2'].astype(object)
        expected[k]['exog_3'] = expected[k]['exog_3'].astype(float)
    
    msg = (
        "Exogenous variables for series 'A' are incomplete. NaNs have been introduced "
        "after setting the frequency."
    )
    with pytest.warns(MissingValuesWarning, match=msg):
        results = reshape_exog_long_to_dict(
            data=data,
            series_id='series_id',
            index='datetime',
            freq='D',
            drop_all_nan_cols=False,
            consolidate_dtypes=True,
        )

    for k in expected.keys():
        pd.testing.assert_frame_equal(results[k], expected[k])


def test_reshape_exog_long_to_dict_output_when_npnan_are_added_in_integer_columns_and_consolidate_true():
    """
    Test the output of the function reshape_exog_long_to_dict when np.nan are added in integer columns
    these columns should be converted to float
    """
    exog_series_1 = pd.DataFrame({
        'exog_1': np.random.normal(0, 1, 10),
        'exog_2': np.random.choice(["A", "B", "C"], 10),
        'exog_3': np.random.randint(0, 10, 10),
        'series': 'series_1',
    }, index = pd.date_range(start='1-1-2000', periods=10, freq='D'))
    exog_series_2 = pd.DataFrame({
        'exog_1': np.random.normal(0, 1, 10),
        'exog_2': np.random.choice(["A", "B"], 10),
        'exog_3': np.random.randint(0, 10, 10),
        'series': 'series_2',
    }, index = pd.date_range(start='1-1-2000', periods=10, freq='D'))
    
    exog_long = (
        pd.concat([exog_series_1, exog_series_2], axis=0)
        .reset_index()
        .rename(columns={"index": "datetime"})
    )
    exog_long = exog_long.loc[
        [0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19], :
    ].copy()
    exog_long["exog_2"] = exog_long["exog_2"].astype("category")
    exog_long["series"] = exog_long["series"].astype("category")

    exog_dict = reshape_exog_long_to_dict(
        data=exog_long,
        series_id="series",
        index="datetime",
        freq="D",
        consolidate_dtypes=True,
        suppress_warnings=True,
    )

    pd.testing.assert_series_equal(exog_dict['series_1'].dtypes, exog_dict['series_2'].dtypes)


def test_reshape_exog_long_to_dict_output_when_npnan_are_added_in_integer_columns_and_consolidate_false():
    """
    Test the output of the function reshape_exog_long_to_dict when np.nan are added in integer columns
    these columns should be converted to float
    """
    exog_series_1 = pd.DataFrame({
        'exog_1': np.random.normal(0, 1, 10),
        'exog_2': np.random.choice(["A", "B", "C"], 10),
        'exog_3': np.random.randint(0, 10, 10),
        'series': 'series_1',
    }, index = pd.date_range(start='1-1-2000', periods=10, freq='D'))
    exog_series_2 = pd.DataFrame({
        'exog_1': np.random.normal(0, 1, 10),
        'exog_2': np.random.choice(["A", "B"], 10),
        'exog_3': np.random.randint(0, 10, 10),
        'series': 'series_2',
    }, index = pd.date_range(start='1-1-2000', periods=10, freq='D'))
    
    exog_long = (
        pd.concat([exog_series_1, exog_series_2], axis=0)
        .reset_index()
        .rename(columns={"index": "datetime"})
    )
    exog_long = exog_long.loc[
        [0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19], :
    ].copy()
    exog_long["exog_2"] = exog_long["exog_2"].astype("category")
    exog_long["series"] = exog_long["series"].astype("category")
    
    exog_dict = reshape_exog_long_to_dict(
        data=exog_long,
        series_id="series",
        index="datetime",
        freq="D",
        consolidate_dtypes=False,
        suppress_warnings=True,
    )
   
    assert exog_dict['series_1'].dtypes.astype(str).to_list() == ['float64', 'category', 'float64']

    if platform.system() == 'Windows':
        assert exog_dict['series_2'].dtypes.astype(str).to_list() == ['float64', 'category', 'int32']
    else:
        assert exog_dict['series_2'].dtypes.astype(str).to_list() == ['float64', 'category', 'int64']



# Unit test _backtesting_forecaster Refit
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.linear_model import Ridge
from sklearn.metrics import mean_squared_error
from skforecast.recursive import ForecasterRecursive
from skforecast.recursive import ForecasterRecursiveClassifier
from skforecast.direct import ForecasterDirect
from skforecast.model_selection._split import TimeSeriesFold
from skforecast.model_selection._validation import _backtesting_forecaster
from skforecast.preprocessing import RollingFeatures

# Fixtures
from skforecast.exceptions import IgnoredArgumentWarning
from ..fixtures_model_selection import y, y_clf
from ..fixtures_model_selection import exog
from ..fixtures_model_selection import out_sample_residuals


# ******************************************************************************
# * Test _backtesting_forecaster No Interval                                   *
# ******************************************************************************

@pytest.mark.parametrize("n_jobs", [-1, 1, 'auto'],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_forecaster_no_exog_no_remainder_ForecasterRecursive_with_mocked(n_jobs):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    ForecasterRecursive.
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.06598802629306816]})
    expected_predictions = pd.DataFrame({
        'pred': np.array([0.55717779, 0.43355138, 0.54969767, 0.52945466, 
                         0.38969292, 0.52778339, 0.49152015, 0.4841678, 
                         0.4076433, 0.50904672, 0.50249462, 0.49232817])}, 
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        n_jobs     = n_jobs,
                                        verbose    = False
                                   )
                                   
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("n_jobs", [-1, 1, 'auto'],
                         ids=lambda n: f'n_jobs: {n}')
def test_output_backtesting_forecaster_ForecasterRecursiveClassifier_with_mocked(n_jobs):
    """
    Test output of _backtesting_forecaster with ForecasterRecursiveClassifier.
    """
    expected_metrics = pd.DataFrame(
        data=[[0.5, 0.45]],
        columns=['accuracy_score', 'balanced_accuracy_score']
    )
    expected_predictions = pd.DataFrame(
        {
            "fold": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2],
            "pred": [
                "bus",
                "train",
                "bus",
                "train",
                "bus",
                "train",
                "bus",
                "train",
                "train",
                "bus",
                "bus",
                "bus",
            ],
            "bus_proba": [
                0.6559317845826198,
                0.3304083470592666,
                0.6559317845826198,
                0.3304083470592666,
                0.763135541754904,
                0.2350409556831433,
                0.763135541754904,
                0.2350409556831433,
                0.3681159967071788,
                0.5967207152844268,
                0.5458169684882528,
                0.5392794662359854,
            ],
            "car_proba": [
                0.30328326075516504,
                0.06028875855870285,
                0.30328326075516504,
                0.06028875855870285,
                0.2098200372652447,
                0.0555164631400065,
                0.2098200372652447,
                0.0555164631400065,
                0.03769250888413279,
                0.09736864130547827,
                0.12853702119356106,
                0.37362423261547634,
            ],
            "train_proba": [
                0.04078495466221505,
                0.6093028943820306,
                0.04078495466221505,
                0.6093028943820306,
                0.02704442097985128,
                0.7094425811768502,
                0.02704442097985128,
                0.7094425811768502,
                0.5941914944086883,
                0.30591064341009494,
                0.32564601031818624,
                0.08709630114853824,
            ],
        },
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )

    forecaster = ForecasterRecursiveClassifier(
        estimator=LogisticRegression(), lags=3
    )

    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_clf) - 12,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y_clf,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = ['accuracy_score', 'balanced_accuracy_score'],
                                        n_jobs     = n_jobs,
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metrics, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_exog_no_remainder_ForecasterDirect_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    ForecasterDirect.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07076203468824617]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.5468482,
                    0.44670961,
                    0.57651222,
                    0.52511275,
                    0.3686309,
                    0.56234835,
                    0.44276032,
                    0.52260065,
                    0.37665741,
                    0.5382938,
                    0.48755548,
                    0.44534071,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterDirect(
                     estimator = LinearRegression(), 
                     lags      = 3,
                     steps     = 4
                 )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        cv         = cv,
                                        exog       = None,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_no_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error'
    """

    expected_metric = pd.DataFrame({"mean_squared_error": [0.06916732087926723]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.48308861,
                    0.5096801,
                    0.49519677,
                    0.47997916,
                    0.49177914,
                    0.495797,
                    0.57738724,
                    0.44370472,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """

    expected_metric = pd.DataFrame({"mean_squared_error": [0.05663345135204598]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.42295275,
                    0.46286083,
                    0.43618422,
                    0.43552906,
                    0.48687517,
                    0.55455072,
                    0.55577332,
                    0.53943402,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_ForecasterRecursive_window_features_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_absolute_error'
    and window features.
    """

    expected_metric = pd.DataFrame({"mean_absolute_error": [0.2095996273]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.4937106691, 0.4471145812, 0.4937808606, 0.536444821 , 0.4215610015,
                    0.4273722215, 0.4483843054, 0.5401413533, 0.4366510863, 0.4705228766,
                    0.5736477861, 0.5938840872,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterRecursive(
        estimator=Ridge(random_state=123), lags=3, window_features=window_features
    )

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_absolute_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_ForecasterDirect_window_features_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_absolute_error'
    and window features.
    """

    expected_metric = pd.DataFrame({"mean_absolute_error": [0.1979777165]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.4754121497, 0.4737280131, 0.5951416701, 0.4792250046, 0.4086254462,
                    0.4906131377, 0.4283798785, 0.4534338932, 0.4384145046, 0.5073137847,
                    0.5507396524, 0.4885913459,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    window_features = RollingFeatures(
        stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation'],
        window_sizes = 3,
    )
    forecaster = ForecasterDirect(
        estimator=Ridge(random_state=123), steps=4, lags=3, window_features=window_features
    )

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_absolute_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.061723961096013524]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.50035119,
                    0.43595809,
                    0.4349167,
                    0.42381237,
                    0.55165332,
                    0.53442833,
                    0.65361802,
                    0.51297419,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster         = forecaster,
                                        y                  = y,
                                        exog               = exog,
                                        cv                 = cv,
                                        metric             = 'mean_squared_error',
                                        verbose            = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_yes_remainder_skip_folds_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked,
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    skip_folds=2
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.04512295747656866]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.50035119,
                    0.65361802,
                    0.51297419,
                ]
            )
        },
        index=[38, 39, 40, 41, 42, 48, 49],
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = 2,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_yes_exog_yes_remainder_skip_folds_intermittent_refit_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked,
    24 observations to backtest, steps=3 (0 remainder), metric='mean_squared_error',
    skip_folds=2 and intermittent refit.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.055092292428684]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array([
                0.43818148, 0.67328031, 0.61839117, 0.61725831, 0.49333332,
                0.39198592, 0.6550626, 0.48837405, 0.54686219, 0.42804696,
                0.4227273, 0.5499255])
        },
        index=[26, 27, 28, 32, 33, 34, 38, 39, 40, 44, 45, 46],
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 2, 2, 2, 4, 4, 4, 6, 6, 6])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 24
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = 2,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster = forecaster,
                                       y          = y,
                                       cv         = cv,
                                       exog       = exog,
                                       metric     = 'mean_squared_error',
                                       verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Test _backtesting_forecaster Interval                                      *
# ******************************************************************************

def test_output_backtesting_forecaster_interval_no_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06598802629306816]})
    expected_predictions = pd.DataFrame(
        data=np.array([
            [0.55717779, 0.19882822, 0.87939731],
            [0.43355138, 0.08976463, 0.86652665],
            [0.54969767, 0.20849403, 0.98096724],
            [0.52945466, 0.16705061, 0.92717139],
            [0.38969292, 0.00644535, 0.745954  ],
            [0.52778339, 0.10488417, 0.96935854],
            [0.49152015, 0.08597044, 0.88803952],
            [0.4841678 , 0.05071879, 0.86119331],
            [0.4076433 , 0.04550155, 0.80099612],
            [0.50904672, 0.11705728, 0.90780612],
            [0.50249462, 0.11010132, 0.89276102],
            [0.49232817, 0.08309367, 0.88100246]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
                             
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_no_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.06916732087926723]})
    expected_predictions = pd.DataFrame(
        data=np.array([
                [0.55717779, 0.19882822, 0.87939731],
                [0.43355138, 0.08976463, 0.86652665],
                [0.54969767, 0.20849403, 0.98096724],
                [0.52945466, 0.16705061, 0.92717139],
                [0.48308861, 0.12658912, 0.89058516],
                [0.5096801 , 0.13148094, 0.90626688],
                [0.49519677, 0.0839043 , 0.92921977],
                [0.47997916, 0.10760776, 0.87837091],
                [0.49177914, 0.07206045, 0.88811027],
                [0.495797  , 0.10849187, 0.89739397],
                [0.57738724, 0.19244588, 0.96037185],
                [0.44370472, 0.0953969 , 0.87177952]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.05663345135204598]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.59059622, 0.24316567, 0.93342085],
            [0.47257504, 0.12485033, 0.87643834],
            [0.53024098, 0.18273747, 0.93078198],
            [0.46163343, 0.10256219, 0.83853371],
            [0.42295275, 0.07085204, 0.73809303],
            [0.46286083, 0.08540127, 0.87827064],
            [0.43618422, 0.04736254, 0.82623837],
            [0.43552906, 0.04081434, 0.77182906],
            [0.48687517, 0.14867238, 0.89104985],
            [0.55455072, 0.19891356, 0.97308891],
            [0.55577332, 0.18085496, 0.96640047],
            [0.53943402, 0.15412602, 0.91288135]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.061723961096013524]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.59059622, 0.24316567, 0.93342085],
                [0.47257504, 0.12485033, 0.87643834],
                [0.53024098, 0.18273747, 0.93078198],
                [0.46163343, 0.10256219, 0.83853371],
                [0.50035119, 0.14089436, 0.88160183],
                [0.43595809, 0.09039821, 0.83346131],
                [0.4349167 , 0.05972659, 0.8553147 ],
                [0.42381237, 0.07678648, 0.80547207],
                [0.55165332, 0.16975842, 0.90196617],
                [0.53442833, 0.16187912, 0.93759341],
                [0.65361802, 0.29167326, 1.07608747],
                [0.51297419, 0.17337054, 0.94324343]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3, binner_kwargs={'n_bins': 15})
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("initial_train_size", 
                         [len(y) - 20, "2022-01-30 00:00:00"],
                         ids=lambda init: f'initial_train_size: {init}')
def test_output_backtesting_forecaster_refit_interval_percentiles_yes_exog(initial_train_size):
    """
    Test output of _backtesting_forecaster with predicted intervals as percentiles.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({"mean_absolute_error": [0.20679065798842525]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55538506, 0.23533636, 0.50028437, 0.69249716, 0.72025328],
            [0.44124674, 0.16831345, 0.39800979, 0.60410175, 0.69095472],
            [0.49752366, 0.2185989 , 0.41366072, 0.65996862, 0.75376276],
            [0.53822797, 0.22537206, 0.44951475, 0.68670224, 0.78470445],
            [0.4793552 , 0.21306399, 0.41434084, 0.65686273, 0.71030187],
            [0.43665118, 0.16671114, 0.41670478, 0.63655954, 0.68152673],
            [0.45763212, 0.20223276, 0.39855006, 0.67033012, 0.71847661],
            [0.4460372 , 0.15349544, 0.38435312, 0.63592087, 0.7011423 ],
            [0.44402791, 0.16958433, 0.37989304, 0.63189191, 0.70047572],
            [0.44985299, 0.18025145, 0.39037074, 0.6280014 , 0.6941219 ],
            [0.44647635, 0.15644524, 0.42272545, 0.60470461, 0.74270915],
            [0.50023368, 0.17611892, 0.48756537, 0.68483285, 0.7824702 ],
            [0.5224527 , 0.23479269, 0.50036463, 0.74289617, 0.82815285],
            [0.44896854, 0.16658387, 0.41769798, 0.68002008, 0.76159033],
            [0.45039083, 0.17140117, 0.40892886, 0.64881926, 0.70647654],
            [0.63590572, 0.40884201, 0.58973532, 0.81614733, 0.82871119],
            [0.6317494 , 0.35642467, 0.58656426, 0.83867971, 0.93000984],
            [0.49592041, 0.25591545, 0.46211508, 0.65788551, 0.80186688],
            [0.52061567, 0.23847963, 0.4387499 , 0.69760195, 0.81649697],
            [0.51002945, 0.25031203, 0.45753096, 0.74133217, 0.80883212]]),
        columns = ['pred', 'p_10', 'p_40', 'p_80', 'p_90'],
        index = pd.date_range(start='2022-01-31', periods=20, freq='D')
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)
    cv = TimeSeriesFold(
             initial_train_size = initial_train_size,
             steps              = 5,
             refit              = True
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y_with_index,
                                       exog                    = exog_with_index,
                                       cv                      = cv,
                                       metric                  = 'mean_absolute_error',
                                       interval                = (10, 40, 80, 90),
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 250,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("interval", 
                         [0.90, (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_output_backtesting_forecaster_interval_conformal_and_binned_with_mocked(interval):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), conformal=True, binned=True.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.064250191230055]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.59059622, 0.2974502 , 0.88374223, 0.89338916,
             0.42635131, 0.31226122, 0.30476807],
            [0.47257504, 0.22356317, 0.72158691, 0.59059622,
             0.89338916, 0.42635131, 0.39818568],
            [0.53024098, 0.14339956, 0.91708239, 0.47257504,
             0.59059622, 0.89338916, 0.70495883],
            [0.46163343, 0.21262156, 0.7106453 , 0.53024098,
             0.47257504, 0.59059622, 0.99535848],
            [0.50035119, 0.35604735, 0.64465502, 0.46163343,
             0.53024098, 0.47257504, 0.35591487],
            [0.41975558, 0.20182965, 0.63768151, 0.31728548,
             0.1156184 , 0.62395295, 0.76254781],
            [0.4256614 , 0.03773832, 0.81358449, 0.41975558,
             0.31728548, 0.1156184 , 0.59317692],
            [0.41176005, 0.19383413, 0.62968598, 0.4256614 ,
             0.41975558, 0.31728548, 0.6917018 ],
            [0.52357817, 0.43392621, 0.61323014, 0.41176005,
             0.4256614 , 0.41975558, 0.15112745],
            [0.509974  , 0.42032204, 0.59962597, 0.52357817,
             0.41176005, 0.4256614 , 0.39887629],
            [0.65354628, 0.29763244, 1.00946013, 0.98555979,
             0.48303426, 0.25045537, 0.2408559 ],
            [0.48210726, 0.17026625, 0.79394828, 0.65354628,
             0.98555979, 0.48303426, 0.34345601]]),
        columns=['pred', 'lower_bound', 'upper_bound', 
                 'lag_1', 'lag_2', 'lag_3', 'exog'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])
    
    forecaster = ForecasterRecursive(
        estimator=LinearRegression(), lags=3, binner_kwargs={'n_bins': 10}
    )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
             steps              = 5,
             initial_train_size = len(y_train),
             refit              = True
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = interval,
                                       interval_method         = 'conformal',
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = True,
                                       random_state            = 123,
                                       return_predictors       = True,
                                       verbose                 = False,
                                       show_progress           = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


@pytest.mark.parametrize("interval", 
                         [0.90, (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_output_backtesting_forecaster_interval_conformal_and_binned_with_mocked_ForecasterDirect(interval):
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    12 observations to backtest, steps=5 (2 remainder), conformal=True, binned=True.
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.063171531991472]})
    expected_predictions = pd.DataFrame(
        data = np.array([
            [0.55584775, 0.37659309, 0.73510242],
            [0.45718425, 0.08077419, 0.83359431],
            [0.57242372, 0.36697393, 0.77787352],
            [0.47236513, 0.10146645, 0.84326382],
            [0.50476389, 0.16013181, 0.84939597],
            [0.4728662 , 0.19898146, 0.74675095],
            [0.44052992, 0.08322808, 0.79783176],
            [0.43209886, 0.07171779, 0.79247994],
            [0.53187641, 0.18347366, 0.88027916],
            [0.54659151, 0.19818876, 0.89499426],
            [0.67106981, 0.31897701, 1.02316261],
            [0.54582075, 0.13585815, 0.95578335]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])
    
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=5, lags=3, binner_kwargs={'n_bins': 10}
    )
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
             steps              = 5,
             initial_train_size = len(y_train),
             refit              = True
         )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster              = forecaster,
                                        y                       = y,
                                        exog                    = exog,
                                        cv                      = cv,
                                        metric                  = 'mean_squared_error',
                                        interval                = interval,
                                        interval_method         = 'conformal',
                                        use_in_sample_residuals = True,
                                        use_binned_residuals    = True,
                                        random_state            = 123,
                                        verbose                 = False,
                                        show_progress           = False
                                   )
    
    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Out sample residuals                                                       *
# ******************************************************************************

def test_output_backtesting_forecaster_interval_out_sample_residuals_no_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    'use_in_sample_residuals = False'
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06598802629306816]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.55717779, 0.63654358, 1.5406994 ],
                [0.43355138, 0.5760887 , 1.53342317],
                [0.54969767, 0.68344933, 1.57809559],
                [0.52945466, 0.69548679, 1.63622589],
                [0.38969292, 0.46905871, 1.37321453],
                [0.52778339, 0.67719181, 1.64080772],
                [0.49152015, 0.65291457, 1.54977963],
                [0.4841678 , 0.63577789, 1.57188267],
                [0.4076433 , 0.48700909, 1.39116491],
                [0.50904672, 0.64496324, 1.58747561],
                [0.50249462, 0.63982498, 1.5237537 ],
                [0.49232817, 0.58731791, 1.50369442]]),
        columns=['pred', 'lower_bound', 'upper_bound'],
        index=pd.RangeIndex(start=38, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)
    forecaster.out_sample_residuals_ = out_sample_residuals
    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = None,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = False,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Callable metric                                                            *
# ******************************************************************************

def my_metric(y_true, y_pred):  # pragma: no cover
    """
    Callable metric
    """
    metric = ((y_true - y_pred) / len(y_true)).mean()
    
    return metric


def test_callable_metric_backtesting_forecaster_no_exog_no_remainder_with_mocked():
    """
    Test callable metric in _backtesting_forecaster with backtesting mocked, interval no. 
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """
    expected_metric = pd.DataFrame({"my_metric": [0.005283745900436151]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.38969292,
                    0.52778339,
                    0.49152015,
                    0.4841678,
                    0.4076433,
                    0.50904672,
                    0.50249462,
                    0.49232817,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster = forecaster,
                                       y          = y,
                                       exog       = None,
                                       cv         = cv,
                                       metric     = my_metric,
                                       verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_list_metrics_backtesting_forecaster_no_exog_no_remainder_with_mocked():
    """
    Test list of metrics in _backtesting_forecaster with backtesting mocked, interval no. 
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error'
    """
    expected_metrics = pd.DataFrame(
        data=[[0.06598802629306816, 0.06598802629306816]],
        columns=['mean_squared_error', 'mean_squared_error']
    )
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.38969292,
                    0.52778339,
                    0.49152015,
                    0.4841678,
                    0.4076433,
                    0.50904672,
                    0.50249462,
                    0.49232817,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metrics, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = ['mean_squared_error', mean_squared_error],
                                        verbose    = False
                                    )

    pd.testing.assert_frame_equal(expected_metrics, metrics)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * fixed_train_size = True                                                    *
# ******************************************************************************

def test_output_backtesting_forecaster_fixed_train_size_no_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    fixed_train_size=True
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.06720844584333846]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.34597367,
                    0.50223873,
                    0.47833829,
                    0.46082257,
                    0.37810191,
                    0.49508366,
                    0.48808014,
                    0.47323313,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_fixed_train_size_no_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, no exog, 
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    fixed_train_size=True
    """
    expected_metric = pd.DataFrame({"mean_squared_error": [0.07217085374372428]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.55717779,
                    0.43355138,
                    0.54969767,
                    0.52945466,
                    0.48308861,
                    0.4909399,
                    0.47942107,
                    0.46025344,
                    0.46649132,
                    0.47061725,
                    0.57603136,
                    0.41480551,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = None,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_fixed_train_size_yes_exog_no_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked,
    12 observations to backtest, steps=4 (no remainder), metric='mean_squared_error',
    fixed_train_size=True
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.05758244401484334]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.37689967,
                    0.44267729,
                    0.42642836,
                    0.41604275,
                    0.45047245,
                    0.53784704,
                    0.53726274,
                    0.51516772,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_fixed_train_size_yes_exog_yes_remainder_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval no.
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked,
    12 observations to backtest, steps=5 (2 remainder), metric='mean_squared_error',
    fixed_train_size=True
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.06425019123005545]})
    expected_predictions = pd.DataFrame(
        {
            "pred": np.array(
                [
                    0.59059622,
                    0.47257504,
                    0.53024098,
                    0.46163343,
                    0.50035119,
                    0.41975558,
                    0.4256614,
                    0.41176005,
                    0.52357817,
                    0.509974,
                    0.65354628,
                    0.48210726,
                ]
            )
        },
        index=pd.RangeIndex(start=38, stop=50, step=1),
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2])

    forecaster = ForecasterRecursive(estimator=LinearRegression(), lags=3)

    n_backtest = 12
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                        forecaster = forecaster,
                                        y          = y,
                                        exog       = exog,
                                        cv         = cv,
                                        metric     = 'mean_squared_error',
                                        verbose    = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Gap                                                                        *
# ******************************************************************************


def test_output_backtesting_forecaster_interval_yes_exog_yes_remainder_gap_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True'
    """
    expected_metric = pd.DataFrame({'mean_squared_error': [0.0839045861490063]})
    expected_predictions = pd.DataFrame(
        data = np.array([[0.65022999,  0.32111688,  1.0092382 ,  0.63097612,
                          0.29371405,  0.22826323,  0.38895057],
                         [0.52364637,  0.17891138,  0.88265459,  0.63097612,
                          0.29371405,  0.22826323,  0.92513249],
                         [0.46809256,  0.14156566,  0.82710077,  0.63097612,
                          0.29371405,  0.22826323,  0.84167   ],
                         [0.48759732,  0.16107043,  0.84202181,  0.63097612,
                          0.29371405,  0.22826323,  0.35739757],
                         [0.47172445,  0.1418827 ,  0.82145924,  0.63097612,
                          0.29371405,  0.22826323,  0.04359146],
                         [0.54075007,  0.20683169,  0.8966551 ,  0.42583029,
                          0.4936851 ,  0.43086276,  0.30476807],
                         [0.50283999,  0.19947461,  0.85291605,  0.42583029,
                          0.4936851 ,  0.43086276,  0.39818568],
                         [0.49737535,  0.17669038,  0.84772905,  0.42583029,
                          0.4936851 ,  0.43086276,  0.70495883],
                         [0.49185456,  0.15793618,  0.83986246,  0.42583029,
                          0.4936851 ,  0.43086276,  0.99535848],
                         [0.48906044,  0.14530175,  0.8391365 ,  0.42583029,
                          0.4936851 ,  0.43086276,  0.35591487],
                         [0.27751064, -0.02499172,  0.67524686,  0.50183668,
                          0.94416002,  0.89338916,  0.76254781],
                         [0.25859617, -0.09051231,  0.59235376,  0.50183668,
                          0.94416002,  0.89338916,  0.59317692],
                         [0.32853669,  0.01962394,  0.6910329 ,  0.50183668,
                          0.94416002,  0.89338916,  0.6917018 ],
                         [0.43751884,  0.13444931,  0.75663214,  0.50183668,
                          0.94416002,  0.89338916,  0.15112745],
                         [0.33016371,  0.01926221,  0.69265992,  0.50183668,
                          0.94416002,  0.89338916,  0.39887629],
                         [0.58126262,  0.21945267,  0.97135055,  0.86630916,
                          0.41482621,  0.31728548,  0.2408559 ],
                         [0.52955296,  0.16642539,  0.94752912,  0.86630916,
                          0.41482621,  0.31728548,  0.34345601]]),
        columns = ['pred', 'lower_bound', 'upper_bound', 
                   'lag_1', 'lag_2', 'lag_3', 'exog'],
        index = pd.RangeIndex(start=33, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3])

    forecaster = ForecasterDirect(
                     estimator = LinearRegression(), 
                     lags      = 3,
                     steps     = 8
                 )
    n_backtest = 20
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = False,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       random_state            = 123,
                                       return_predictors       = True,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_interval_yes_exog_not_allow_remainder_gap_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True', allow_incomplete_fold = False
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({'mean_squared_error': [0.09133694038363274]})
    expected_predictions = pd.DataFrame(
        data = np.array([[ 0.65022999,  0.32111688,  1.0092382 ],
                        [ 0.52364637,  0.17891138,  0.88265459],
                        [ 0.46809256,  0.14156566,  0.82710077],
                        [ 0.48759732,  0.16107043,  0.84202181],
                        [ 0.47172445,  0.1418827 ,  0.82145924],
                        [ 0.50952084,  0.18013674,  0.83879308],
                        [ 0.52131987,  0.19865195,  0.82558503],
                        [ 0.50289964,  0.18029903,  0.7872203 ],
                        [ 0.54941988,  0.22003578,  0.81990615],
                        [ 0.51121195,  0.18536794,  0.84048419],
                        [ 0.25123452, -0.05173225,  0.58626359],
                        [ 0.2791409 , -0.01997518,  0.6141032 ],
                        [ 0.28390161, -0.02663537,  0.60172032],
                        [ 0.38317935,  0.09272498,  0.72647866],
                        [ 0.42183128,  0.11886451,  0.73965   ]]),
        columns = ['pred', 'lower_bound', 'upper_bound'],
        index = pd.date_range(start='2022-02-03', periods=15, freq='D')
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2])

    forecaster = ForecasterDirect(
                     estimator = LinearRegression(), 
                     lags      = 3,
                     steps     = 8
                 )
    cv = TimeSeriesFold(
            steps                 = 5,
            initial_train_size    = len(y_with_index) - 20,
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y_with_index,
                                       exog                    = exog_with_index,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


# ******************************************************************************
# * Refit int                                                                  *
# ******************************************************************************

def test_output_backtesting_forecaster_refit_int_interval_yes_exog_yes_remainder_with_mocked():
    """
    Test output of backtesting_forecaster refit with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=0, metric='mean_squared_error',
    'use_in_sample_residuals = True'. Refit int.
    """
    expected_metric = pd.DataFrame(
        {'mean_squared_error': [0.06099110404144631], 
         'mean_absolute_scaled_error': [0.791393]}
    )
    expected_predictions = pd.DataFrame(
        data = np.array([[0.55616986, 0.20571296, 0.92594217],
                        [0.48751797, 0.12416328, 0.85598316],
                        [0.57764391, 0.22718701, 0.94741621],
                        [0.51298667, 0.14963198, 0.88145186],
                        [0.47430051, 0.1506004 , 0.8143182 ],
                        [0.49192271, 0.16822259, 0.8319404 ],
                        [0.52213783, 0.19843771, 0.86215552],
                        [0.54492575, 0.22122564, 0.88494344],
                        [0.52501537, 0.20809452, 0.89438955],
                        [0.4680474 , 0.1611217 , 0.78353678],
                        [0.51059498, 0.19367413, 0.87996916],
                        [0.53067132, 0.22374562, 0.84616071],
                        [0.4430938 , 0.09119872, 0.80766626],
                        [0.49911716, 0.14722208, 0.8776036 ],
                        [0.44546347, 0.09356839, 0.81003593],
                        [0.46530749, 0.11341241, 0.84379393],
                        [0.46901878, 0.10970301, 0.8566689 ],
                        [0.55371362, 0.21839276, 0.96008583],
                        [0.60759064, 0.24827486, 0.99524075],
                        [0.50415336, 0.16883251, 0.91052558]]),
        columns = ['pred', 'lower_bound', 'upper_bound'],
        index = pd.RangeIndex(start=30, stop=50, step=1)
    )
    expected_predictions.insert(0, 'fold', [0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9])

    forecaster = ForecasterDirect(
                     estimator = Ridge(random_state=123), 
                     lags      = 3,
                     steps     = 8
                 )

    n_backtest = 20
    y_train = y[:-n_backtest]
    cv = TimeSeriesFold(
            steps                 = 2,
            initial_train_size    = len(y_train),
            window_size           = None,
            differentiation       = None,
            refit                 = 2,
            fixed_train_size      = False,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = ['mean_squared_error', 'mean_absolute_scaled_error'],
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False,
                                       n_jobs                  = 1
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_refit_int_interval_yes_exog_not_allow_remainder_gap_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True', allow_incomplete_fold = False. Refit int.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({'mean_squared_error': [0.060991643719298785]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.51878642, 0.17536117, 0.88612664],
                [0.49269791, 0.14299038, 0.83669243],
                [0.49380441, 0.15379905, 0.85842322],
                [0.51467463, 0.16024582, 0.94426076],
                [0.52690045, 0.1834752 , 0.89424067],
                [0.51731996, 0.16761242, 0.86131448],
                [0.51290311, 0.17289775, 0.87752193],
                [0.50334306, 0.14891425, 0.93292919],
                [0.50171526, 0.15829001, 0.86905548],
                [0.50946908, 0.15976155, 0.8534636 ],
                [0.50200357, 0.16199821, 0.86662238],
                [0.50436041, 0.1499316 , 0.93394654],
                [0.4534189 , 0.05348442, 0.83437551],
                [0.52621695, 0.15188405, 0.90769545],
                [0.50477802, 0.13426951, 0.89447609],
                [0.52235258, 0.16382552, 0.91219012]]),
                columns = ['pred', 'lower_bound', 'upper_bound'],
        index = pd.date_range(start='2022-02-03', periods=16, freq='D')
    )
    expected_predictions.insert(0, 'fold', [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3])

    forecaster = ForecasterRecursive(
        estimator=Ridge(random_state=123), lags=3, binner_kwargs={'n_bins': 15}
    )
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_with_index) - 20,
            window_size           = None,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = True,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = False,
        )

    warn_msg = re.escape(
        "If `refit` is an integer other than 1 (intermittent refit). `n_jobs` "
        "is set to 1 to avoid unexpected results during parallelization."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        metric, backtest_predictions = _backtesting_forecaster(
                                           forecaster              = forecaster,
                                           y                       = y_with_index,
                                           exog                    = exog_with_index,
                                           cv                      = cv,
                                           metric                  = 'mean_squared_error',
                                           interval                = [5, 95],
                                           interval_method         = 'bootstrapping',
                                           n_boot                  = 500,
                                           random_state            = 123,
                                           use_in_sample_residuals = True,
                                           use_binned_residuals    = False,
                                           verbose                 = False,
                                           n_jobs                  = 2
                                       )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_refit_int_interval_yes_exog_fold_stride_skip_folds_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=10 and gap=0, metric='mean_squared_error',
    'use_in_sample_residuals = True', allow_incomplete_fold = True. Refit, 
    fold_stride 2 and skip_folds=[2].
    """

    expected_metric = pd.DataFrame({'mean_squared_error': [0.066569074390952]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.        , 0.53190531, 0.20645634, 0.86696728],
                [0.        , 0.47614301, 0.11951144, 0.79059385],
                [0.        , 0.51048481, 0.16460017, 0.87458166],
                [0.        , 0.51878642, 0.17536117, 0.88612664],
                [0.        , 0.49269791, 0.14299038, 0.83669243],
                [0.        , 0.49380441, 0.15379905, 0.85842322],
                [0.        , 0.51467463, 0.16024582, 0.94426076],
                [0.        , 0.52689128, 0.16464806, 0.89120127],
                [0.        , 0.51740207, 0.19822415, 0.92131869],
                [0.        , 0.51314964, 0.18154424, 0.87347311],
                [1.        , 0.46808478, 0.10221064, 0.74574734],
                [1.        , 0.45857207, 0.11762277, 0.74337719],
                [1.        , 0.45965376, 0.10546415, 0.76994217],
                [1.        , 0.4537546 , 0.08662693, 0.76687432],
                [1.        , 0.45522748, 0.10384957, 0.76570134],
                [1.        , 0.45718251, 0.09019411, 0.76484898],
                [1.        , 0.46065319, 0.10487234, 0.77548387],
                [1.        , 0.46486462, 0.10160111, 0.77794994],
                [1.        , 0.4571576 , 0.11694018, 0.76226993],
                [1.        , 0.46140341, 0.09992819, 0.7517024 ],
                [3.        , 0.44281847, 0.14097126, 0.8113343 ],
                [3.        , 0.44603831, 0.06778012, 0.80274524],
                [3.        , 0.45849586, 0.09241489, 0.84740755],
                [3.        , 0.4534189 , 0.05348442, 0.83437551],
                [3.        , 0.52621695, 0.15188405, 0.90769545],
                [3.        , 0.50477802, 0.13426951, 0.89447609],
                [3.        , 0.52235258, 0.16382552, 0.91219012],
                [3.        , 0.51277574, 0.16827985, 0.89130533],
                [4.        , 0.52962072, 0.07588641, 0.89958312],
                [4.        , 0.45235044, 0.10297558, 0.83460183],
                [4.        , 0.54318556, 0.17324367, 0.92057939],
                [4.        , 0.50977045, 0.19361858, 0.88735246]]),
                columns = ['fold', 'pred', 'lower_bound', 'upper_bound'],
        index = pd.Index(
                    [30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 34, 35, 36, 37, 38, 39, 40, 41,
                     42, 43, 42, 43, 44, 45, 46, 47, 48, 49, 46, 47, 48, 49]
                )
    ).astype({'fold': int})

    forecaster = ForecasterRecursive(
        estimator=Ridge(random_state=123), lags=3, binner_kwargs={'n_bins': 15}
    )
    cv = TimeSeriesFold(
            steps                 = 10,
            initial_train_size    = len(y) - 20,
            fold_stride           = 4,
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = [2],
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )

    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster              = forecaster,
                                       y                       = y,
                                       exog                    = exog,
                                       cv                      = cv,
                                       metric                  = 'mean_squared_error',
                                       interval                = [5, 95],
                                       interval_method         = 'bootstrapping',
                                       n_boot                  = 500,
                                       random_state            = 123,
                                       use_in_sample_residuals = True,
                                       use_binned_residuals    = False,
                                       verbose                 = False,
                                       n_jobs                  = 2
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_refit_int_interval_yes_exog_not_allow_remainder_gap_fold_stride_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked, interval yes. 
    Estimator is LinearRegression with lags=3, Series y is mocked, exog is mocked, 
    20 observations to backtest, steps=5 and gap=3, metric='mean_squared_error',
    'use_in_sample_residuals = True', allow_incomplete_fold = False. Refit int and
    fold_stride 2.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({'mean_squared_error': [0.061008068987864]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.        , 0.51878642, 0.17536117, 0.88612664],
                [0.        , 0.49269791, 0.14299038, 0.83669243],
                [0.        , 0.49380441, 0.15379905, 0.85842322],
                [0.        , 0.51467463, 0.16024582, 0.94426076],
                [1.        , 0.49776158, 0.15433633, 0.8651018 ],
                [1.        , 0.51407679, 0.16436926, 0.85807131],
                [1.        , 0.52547527, 0.1854699 , 0.89009408],
                [1.        , 0.51757577, 0.16314696, 0.9471619 ],
                [2.        , 0.52690045, 0.1834752 , 0.89424067],
                [2.        , 0.51731996, 0.16761242, 0.86131448],
                [2.        , 0.51290311, 0.17289775, 0.87752193],
                [2.        , 0.50334306, 0.14891425, 0.93292919],
                [3.        , 0.44432915, 0.08884275, 0.72905021],
                [3.        , 0.44322737, 0.12273286, 0.75351458],
                [3.        , 0.435265  , 0.07465634, 0.75224648],
                [3.        , 0.44653356, 0.10638146, 0.75176957],
                [4.        , 0.44545297, 0.08996657, 0.73017403],
                [4.        , 0.44229531, 0.12180081, 0.75258253],
                [4.        , 0.44318094, 0.08257228, 0.76016242],
                [4.        , 0.44331663, 0.10316454, 0.74855265],
                [5.        , 0.43963505, 0.08414865, 0.72435611],
                [5.        , 0.44536465, 0.12487014, 0.75565186],
                [5.        , 0.44300857, 0.08239991, 0.75999004],
                [5.        , 0.45037258, 0.11022049, 0.75560859],
                [6.        , 0.4534189 , 0.05348442, 0.83437551],
                [6.        , 0.52621695, 0.15188405, 0.90769545],
                [6.        , 0.50477802, 0.13426951, 0.89447609],
                [6.        , 0.52235258, 0.16382552, 0.91219012]]),
                columns = ['fold', 'pred', 'lower_bound', 'upper_bound'],
        index = pd.DatetimeIndex(
                    ['2022-02-03', '2022-02-04', '2022-02-05', '2022-02-06',
                    '2022-02-05', '2022-02-06', '2022-02-07', '2022-02-08',
                    '2022-02-07', '2022-02-08', '2022-02-09', '2022-02-10',
                    '2022-02-09', '2022-02-10', '2022-02-11', '2022-02-12',
                    '2022-02-11', '2022-02-12', '2022-02-13', '2022-02-14',
                    '2022-02-13', '2022-02-14', '2022-02-15', '2022-02-16',
                    '2022-02-15', '2022-02-16', '2022-02-17', '2022-02-18']
                )
    ).astype({'fold': int})

    forecaster = ForecasterRecursive(
        estimator=Ridge(random_state=123), lags=3, binner_kwargs={'n_bins': 15}
    )
    cv = TimeSeriesFold(
            steps                 = 4,
            initial_train_size    = len(y_with_index) - 20,
            fold_stride           = 2,
            window_size           = None,
            differentiation       = None,
            refit                 = 3,
            fixed_train_size      = True,
            gap                   = 3,
            skip_folds            = None,
            allow_incomplete_fold = False,
            return_all_indexes    = False,
        )

    warn_msg = re.escape(
        "If `refit` is an integer other than 1 (intermittent refit). `n_jobs` "
        "is set to 1 to avoid unexpected results during parallelization."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        metric, backtest_predictions = _backtesting_forecaster(
                                           forecaster              = forecaster,
                                           y                       = y_with_index,
                                           exog                    = exog_with_index,
                                           cv                      = cv,
                                           metric                  = 'mean_squared_error',
                                           interval                = [5, 95],
                                           interval_method         = 'bootstrapping',
                                           n_boot                  = 500,
                                           random_state            = 123,
                                           use_in_sample_residuals = True,
                                           use_binned_residuals    = False,
                                           verbose                 = False,
                                           n_jobs                  = 2
                                       )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)


def test_output_backtesting_forecaster_refit_exog_fold_stride_greater_than_steps_with_mocked():
    """
    Test output of _backtesting_forecaster with backtesting mocked when fold_stride
    is greater than steps.
    """
    y_with_index = y.copy()
    y_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    exog_with_index = exog.copy()
    exog_with_index.index = pd.date_range(start='2022-01-01', periods=50, freq='D')

    expected_metric = pd.DataFrame({'mean_squared_error': [0.077634587544307]})
    expected_predictions = pd.DataFrame(
        data = np.array([
                [0.        , 0.53898423],
                [0.        , 0.48834955],
                [0.        , 0.57145241],
                [1.        , 0.44306238],
                [1.        , 0.45703358],
                [1.        , 0.43672228],
                [2.        , 0.43330016],
                [2.        , 0.48078074],
                [2.        , 0.47218491],
                [3.        , 0.5986394 ],
                [3.        , 0.55012333]]),
                columns = ['fold', 'pred'],
        index = pd.DatetimeIndex(
                    ['2022-01-31', '2022-02-01', '2022-02-02', '2022-02-06',
                     '2022-02-07', '2022-02-08', '2022-02-12', '2022-02-13',
                     '2022-02-14', '2022-02-18', '2022-02-19']
                )
    ).astype({'fold': int})

    forecaster = ForecasterDirect(
        estimator=Ridge(random_state=123), steps=3, lags=3
    )
    cv = TimeSeriesFold(
            steps                 = 3,
            initial_train_size    = len(y_with_index) - 20,
            fold_stride           = 6,
            window_size           = None,
            differentiation       = None,
            refit                 = True,
            fixed_train_size      = True,
            gap                   = 0,
            skip_folds            = None,
            allow_incomplete_fold = True,
            return_all_indexes    = False,
        )
    
    metric, backtest_predictions = _backtesting_forecaster(
                                       forecaster   = forecaster,
                                       y            = y_with_index,
                                       exog         = exog_with_index,
                                       cv           = cv,
                                       metric       = 'mean_squared_error',
                                       interval     = None,
                                       random_state = 123,
                                       verbose      = True,
                                   )

    pd.testing.assert_frame_equal(expected_metric, metric)
    pd.testing.assert_frame_equal(expected_predictions, backtest_predictions)



# Unit test fit ForecasterRecursiveMultiSeries
# ==============================================================================
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from skforecast.preprocessing import RollingFeatures
from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_dict_range,
    exog_wide_range,
    exog_dict_range
)

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_forecaster_series_exog_features_stored():
    """
    Test forecaster stores series and exog features after fitting.
    """
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, window_features=rolling, transformer_exog=transformer_exog
    )
    forecaster.fit(series=series_dict_range, exog=exog_wide_range)

    series_names_in_ = ['l1', 'l2']
    exog_in_ = True
    exog_type_in_ = type(exog_wide_range)
    exog_names_in_ = ['exog_1', 'exog_2']
    exog_dtypes_in_ = {
        'exog_1': np.dtype(float), 
        'exog_2': np.dtype(object)
    }
    # All floats
    exog_dtypes_out_ = {
        'exog_1': np.dtype(float), 
        'exog_2_a': np.dtype(float), 
        'exog_2_b': np.dtype(float)
    }
    X_train_series_names_in_ = ['l1', 'l2']
    X_train_window_features_names_out_ = ['roll_ratio_min_max_4', 'roll_median_4']
    X_train_exog_names_out_ = ['exog_1', 'exog_2_a', 'exog_2_b']
    X_train_features_names_out_ = [
        'lag_1', 'lag_2', 'lag_3', 'roll_ratio_min_max_4', 'roll_median_4', 
        '_level_skforecast', 'exog_1', 'exog_2_a', 'exog_2_b'
    ]
    
    assert forecaster.series_names_in_ == series_names_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_series_names_in_ == X_train_series_names_in_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_


def test_forecaster_series_not_matching_exog_features_stored():
    """
    Test forecaster stores series and exog features after fitting when exog keys
    do not match series keys.
    """
    exog_dict_no_match = {
        '1': exog_dict_range['l1'].copy(),
        '2': exog_dict_range['l2'].copy(),
    }
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, window_features=rolling, transformer_exog=transformer_exog
    )
    forecaster.fit(series=series_dict_range, exog=exog_dict_no_match)

    series_names_in_ = ['l1', 'l2']
    exog_in_ = False
    exog_type_in_ = None
    exog_names_in_ = None
    exog_dtypes_in_ = None
    exog_dtypes_out_ = None
    X_train_series_names_in_ = ['l1', 'l2']
    X_train_window_features_names_out_ = ['roll_ratio_min_max_4', 'roll_median_4']
    X_train_exog_names_out_ = None
    X_train_features_names_out_ = [
        'lag_1', 'lag_2', 'lag_3', 'roll_ratio_min_max_4', 'roll_median_4', '_level_skforecast'
    ]
    
    assert forecaster.series_names_in_ == series_names_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_series_names_in_ == X_train_series_names_in_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_


def test_fit_correct_dict_create_series_weights_weight_func_transformer_series():
    """
    Test fit method creates correctly all the auxiliary dicts, series_weights_,
    weight_func_, transformer_series_.
    """
    series = pd.DataFrame(
        {'l1': np.arange(10), 'l2': np.arange(10), 'l3': np.arange(10)},
        index = pd.date_range('2022-01-04', periods=10, freq='D')
    ).to_dict(orient='series')

    def custom_weights(index):  # pragma: no cover
        """
        Return 0 if index is between '2022-01-08' and '2022-01-10', 1 otherwise.
        """
        weights = np.where(
                      (index >= '2022-01-08') & (index <= '2022-01-10'),
                      0,
                      1
                  )
        
        return weights

    transformer_series = {'l1': StandardScaler(), '_unknown_level': StandardScaler()}
    weight_func = {'l2': custom_weights}    
    series_weights = {'l1': 3., 'l3': 0.5, 'l4': 2.}

    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(), 
                     lags               = 3,
                     transformer_series = transformer_series,
                     weight_func        = weight_func,
                     series_weights     = series_weights
                 )
    
    forecaster.fit(series=series, store_in_sample_residuals=False)

    expected_transformer_series_ = {
        'l1': forecaster.transformer_series_['l1'], 
        'l2': None, 
        'l3': None,
        '_unknown_level': forecaster.transformer_series_['_unknown_level']
    }
    expected_weight_func_ = {
        'l1': forecaster._weight_func_all_1, 
        'l2': custom_weights, 
        'l3': forecaster._weight_func_all_1
    }
    expected_series_weights_ = {'l1': 3., 'l2': 1., 'l3': 0.5}

    assert forecaster.transformer_series_.keys() == expected_transformer_series_.keys()
    for k in expected_transformer_series_.keys():
        if expected_transformer_series_[k] is None:
            assert isinstance(forecaster.transformer_series_[k], type(None))
        else:
            assert isinstance(forecaster.transformer_series_[k], StandardScaler)
    assert forecaster.weight_func_.keys() == expected_weight_func_.keys()
    for key in forecaster.weight_func_.keys():
        assert forecaster.weight_func_[key].__code__.co_code == expected_weight_func_[key].__code__.co_code
    assert forecaster.series_weights_ == expected_series_weights_

    forecaster.fit(
        series = {k: v for k, v in series.items() if k in ['l1', 'l2']}, 
        store_in_sample_residuals = False
    )

    expected_transformer_series_ = {
        'l1': forecaster.transformer_series_['l1'], 
        'l2': None,
        '_unknown_level': forecaster.transformer_series_['_unknown_level']
    }
    expected_weight_func_ = {'l1': forecaster._weight_func_all_1, 'l2': custom_weights}
    expected_series_weights_ = {'l1': 3., 'l2': 1.}

    assert forecaster.transformer_series_.keys() == expected_transformer_series_.keys()
    for k in expected_transformer_series_.keys():
        if expected_transformer_series_[k] is None:
            assert isinstance(forecaster.transformer_series_[k], type(None))
        else:
            assert isinstance(forecaster.transformer_series_[k], StandardScaler)
    assert forecaster.weight_func_.keys() == expected_weight_func_.keys()
    for key in forecaster.weight_func_.keys():
        assert forecaster.weight_func_[key].__code__.co_code == expected_weight_func_[key].__code__.co_code
    assert forecaster.series_weights_ == expected_series_weights_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test serie_with_DatetimeIndex.index.freq is stored in forecaster.index_freq_.
    """
    series = {
        '1': pd.Series(np.arange(5)), 
        '2': pd.Series(np.arange(5))
    }
    series['1'].index = pd.date_range(start='2022-01-01', periods=5, freq='D')
    series['2'].index = pd.date_range(start='2022-01-01', periods=5, freq='D')

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series)
    results = forecaster.index_freq_

    expected = series['1'].index.freq

    assert results == expected


def test_forecaster_index_step_stored():
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq_.
    """
    series = {
        '1': pd.Series(np.arange(5)), 
        '2': pd.Series(np.arange(5))
    }
    
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series)
    results = forecaster.index_freq_

    expected = series['1'].index.step

    assert results == expected


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot'], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_in_sample_residuals_stored(encoding):
    """
    Test that values of in_sample_residuals_ are stored after fitting
    when `store_in_sample_residuals=True`.
    """
    series = {
        '1': pd.Series(np.arange(5)), 
        '2': pd.Series(np.arange(5))
    }

    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, window_features=rolling
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results = forecaster.in_sample_residuals_

    expected = {
        '1': np.array([-4.4408921e-16, 0.0000000e+00]),
        '2': np.array([0., 0.]),
        '_unknown_level': np.array([-4.4408921e-16, 0.0000000e+00, 0., 0.])
    }
    
    X_train_window_features_names_out_ = ['roll_ratio_min_max_4', 'roll_median_4']
    X_train_features_names_out_ = (
        ['lag_1', 'lag_2', 'lag_3', 'roll_ratio_min_max_4', 'roll_median_4', '_level_skforecast'] 
        if encoding != 'onehot' 
        else ['lag_1', 'lag_2', 'lag_3', 'roll_ratio_min_max_4', 'roll_median_4', '1', '2']
    )

    assert forecaster.series_names_in_ == ['1', '2']
    assert forecaster.X_train_series_names_in_ == ['1', '2']
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_
    assert isinstance(results, dict)
    assert np.all(isinstance(x, np.ndarray) for x in results.values())
    assert results.keys() == expected.keys()
    assert np.all(np.all(np.isclose(results[k], expected[k])) for k in expected.keys())


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot', None], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_in_sample_residuals_by_bin_stored(encoding):
    """
    Test that values of in_sample_residuals_ are stored after fitting
    when `store_in_sample_residuals=True`.
    """
    rng = np.random.default_rng(1894)
    series = {
        "1": pd.Series(rng.normal(10, 5, 20)), 
        "2": pd.Series(rng.normal(10, 5, 20))
    }

    rolling = RollingFeatures(stats=["ratio_min_max", "median"], window_sizes=4)
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, window_features=rolling, binner_kwargs={"n_bins": 3}
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_residuals = forecaster.in_sample_residuals_
    results_residuals_bin = forecaster.in_sample_residuals_by_bin_
    results_binner_intervals = forecaster.binner_intervals_

    if encoding is not None:
        expected_residuals = {
            '1': np.array([-1.31688412, -4.83850341, -6.89193651, -3.88484903,  0.93550062,
                    -5.05052676,  2.22146909,  8.10253854, -0.03935851,  2.44731007,
                    0.10143843,  0.99207366,  4.43782989, -6.31608079,  2.76608522,
                    6.3338936 ]),
            '2': np.array([ 1.35912405, -6.75955818, -1.96209665,  1.38750505,  3.0777547 ,
                    2.66627803,  7.45266459, -4.81736063,  2.97073109, -2.40028892,
                    0.69596738,  6.12813429, -2.54249734, -1.75531268, -2.58079634,
                    -2.92024843]),
            '_unknown_level': np.array([-1.31688412, -4.83850341, -6.89193651, -3.88484903,  0.93550062,
                    -5.05052676,  2.22146909,  8.10253854, -0.03935851,  2.44731007,
                    0.10143843,  0.99207366,  4.43782989, -6.31608079,  2.76608522,
                    6.3338936 ,  1.35912405, -6.75955818, -1.96209665,  1.38750505,
                    3.0777547 ,  2.66627803,  7.45266459, -4.81736063,  2.97073109,
                    -2.40028892,  0.69596738,  6.12813429, -2.54249734, -1.75531268,
                    -2.58079634, -2.92024843])
        }

        expected_residuals_bin = {
            '1': {
                    0: np.array([-1.31688412, -4.83850341,  0.99207366,  2.76608522,  6.3338936 ]),
                    1: np.array([-3.88484903,  0.93550062,  2.44731007,  0.10143843,  4.43782989, -6.31608079]),
                    2: np.array([-6.89193651, -5.05052676,  2.22146909,  8.10253854, -0.03935851])
                },
            '2': {
                    0: np.array([-6.75955818, -1.96209665,  3.0777547 ,  2.97073109, -2.40028892]),
                    1: np.array([1.38750505,  7.45266459, -4.81736063,  0.69596738, -2.92024843]),
                    2: np.array([1.35912405,  2.66627803,  6.12813429, -2.54249734, -1.75531268, -2.58079634])
                },
            '_unknown_level': {
                    0: np.array([-1.31688412, -4.83850341, -3.88484903,  0.99207366,  2.76608522,
                            6.3338936, -6.75955818, -1.96209665,  3.0777547,  2.97073109, -2.40028892]),
                    1: np.array([ 0.93550062, -0.03935851,  2.44731007,  0.10143843,  4.43782989,
                            -6.31608079,  7.45266459, -4.81736063,  0.69596738, -2.92024843]),
                    2: np.array([-6.89193651, -5.05052676,  2.22146909,  8.10253854,  1.35912405,
                            1.38750505,  2.66627803,  6.12813429, -2.54249734, -1.75531268, -2.58079634])
                }
            }

        expected_binner_intervals = {
            "1": {
                0: (7.268796921364613, 8.987998904684593),
                1: (8.987998904684593, 9.830020290458247),
                2: (9.830020290458247, 12.366369800747009),
            },
            "2": {
                0: (7.211558912361964, 9.22351564748152),
                1: (9.22351564748152, 11.350695147373349),
                2: (11.350695147373349, 11.844709741478395),
            },
            "_unknown_level": {
                0: (7.211558912361964, 9.016289033159751),
                1: (9.016289033159751, 10.730519759760567),
                2: (10.730519759760567, 12.366369800747009),
            },
        }

    else:
        expected_residuals = {
            '_unknown_level': np.array([-2.05820541, -5.47429818, -7.31524702, -4.29526858,  0.66810725,
                -5.37727919,  1.94669051,  7.67911837, -0.63572989,  1.79063442,
                -0.48881591,  0.44689235,  4.0759107 , -6.74143431,  2.16887587,
                5.7738583 ,  1.76518089, -6.32949235, -1.56689305,  1.98201019,
                3.56844034,  3.14094834,  7.85637887, -4.32241071,  3.27047182,
                -2.01737457,  1.26390611,  6.83575019, -2.02500966, -1.25537261,
                -2.00844951, -2.32189357])
        }

        expected_residuals_bin = {
            '_unknown_level': {
                0: np.array([-2.05820541, -5.47429818,  0.44689235,  2.16887587, -6.32949235,
                            -1.56689305,  3.56844034,  3.27047182, -2.01737457,  1.26390611, -2.32189357]),
                1: np.array([-4.29526858,  0.66810725,  1.79063442, -0.48881591,  4.0759107, -6.74143431,
                            5.7738583,  1.98201019,  7.85637887, -4.32241071]),
                2: np.array([-7.31524702, -5.37727919,  1.94669051,  7.67911837, -0.63572989,
                            1.76518089,  3.14094834,  6.83575019, -2.02500966, -1.25537261, -2.00844951])
            }
        }

        expected_binner_intervals = {
            "_unknown_level": {
                0: (6.81635531697439, 9.303028105789933),
                1: (9.303028105789933, 10.690051434777626),
                2: (10.690051434777626, 12.789680311772173),
            }
        }

    X_train_window_features_names_out_ = ["roll_ratio_min_max_4", "roll_median_4"]
    if encoding in ['ordinal', 'ordinal_category']:
        X_train_features_names_out_ = [
            "lag_1",
            "lag_2",
            "lag_3",
            "roll_ratio_min_max_4",
            "roll_median_4",
            "_level_skforecast",
        ]
    elif encoding == 'onehot':
        X_train_features_names_out_ = ["lag_1", "lag_2", "lag_3", "roll_ratio_min_max_4", "roll_median_4", "1", "2"]
    else:
        X_train_features_names_out_ = ["lag_1", "lag_2", "lag_3", "roll_ratio_min_max_4", "roll_median_4"]

    assert forecaster.series_names_in_ == ["1", "2"]
    assert forecaster.X_train_series_names_in_ == ["1", "2"]
    assert (
        forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    )
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_

    # In-sample residuals
    assert results_residuals.keys() == expected_residuals.keys()
    for k in results_residuals.keys():
        np.testing.assert_array_almost_equal(
            np.sort(results_residuals[k]), np.sort(expected_residuals[k])
        )

    # In-sample residuals by bin
    assert results_residuals_bin.keys() == expected_residuals_bin.keys()
    for level in results_residuals_bin.keys():
        assert results_residuals_bin[level].keys() == expected_residuals_bin[level].keys()
        for k in results_residuals_bin[level].keys():
            np.testing.assert_array_almost_equal(
                results_residuals_bin[level][k], expected_residuals_bin[level][k]
            )
    
    # Binner intervals
    assert results_binner_intervals.keys() == expected_binner_intervals.keys()
    for level in results_binner_intervals.keys():
        assert results_binner_intervals[level].keys() == expected_binner_intervals[level].keys()
        for k in results_binner_intervals[level].keys():
            assert results_binner_intervals[level][k][0] == approx(expected_binner_intervals[level][k][0])
            assert results_binner_intervals[level][k][1] == approx(expected_binner_intervals[level][k][1])


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot', None], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_same_residuals_when_residuals_greater_than_10_000(encoding):
    """
    Test fit return same residuals when residuals len is greater than 10_000.
    Testing with two different forecaster. Residuals shouldn't be more than
    1000 values.
    """
    rng = np.random.default_rng(12345)
    series = pd.DataFrame(
        {"1": rng.normal(10, 1, 15_000), "2": rng.normal(10, 1, 15_000)},
        index=pd.date_range(start="2000-01-01", periods=15_000, freq="h"),
    ).to_dict(orient='series')

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_1 = forecaster.in_sample_residuals_

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_2 = forecaster.in_sample_residuals_

    assert results_1.keys() == results_2.keys()
    assert np.all([len(v) == 10_000 for v in results_1.values()])
    assert np.all([len(v) == 10_000 for v in results_2.values()])
    assert np.all(np.all(results_1[k] == results_2[k]) for k in results_2.keys())


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot', None], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_same_residuals_by_bin_when_residuals_greater_than_10_000(encoding):
    """
    Test fit return same residuals by bien when residuals len is greater than 10_000.
    Testing with two different forecaster. Residuals shouldn't be more than
    1000 values.
    """
    rng = np.random.default_rng(12345)
    series = pd.DataFrame(
        {"1": rng.normal(10, 1, 15_000), "2": rng.normal(10, 1, 15_000)},
        index=pd.date_range(start="2000-01-01", periods=15_000, freq="h"),
    ).to_dict(orient='series')

    forecaster_1 = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={"n_bins": 3}
    )
    forecaster_1.fit(series=series, store_in_sample_residuals=True)
    results_1 = forecaster_1.in_sample_residuals_by_bin_

    forecaster_2 = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={"n_bins": 3}
    )
    forecaster_2.fit(series=series, store_in_sample_residuals=True)
    results_2 = forecaster_2.in_sample_residuals_by_bin_

    assert results_1.keys() == results_2.keys()
    for k in results_1.keys() & results_2.keys():
        for bin in results_1[k].keys():
            assert len(results_1[k][bin]) == len(results_2[k][bin])
            assert len(results_1[k][bin]) == 10_000 // 3
    
    assert forecaster_1.binner_intervals_.keys() == forecaster_2.binner_intervals_.keys()
    for level in forecaster_1.binner_intervals_.keys():
        for k in forecaster_1.binner_intervals_[level].keys():
            assert forecaster_1.binner_intervals_[level][k][0] == forecaster_2.binner_intervals_[level][k][0]
            assert forecaster_1.binner_intervals_[level][k][1] == forecaster_2.binner_intervals_[level][k][1]


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot', None], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_in_sample_residuals_not_stored_probabilistic_mode_binned(encoding):
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False`. Binner intervals are stored.
    """
    rng = np.random.default_rng(1894)
    series = {
        "1": pd.Series(rng.normal(10, 5, 20)), 
        "2": pd.Series(rng.normal(10, 5, 20))
    }

    rolling = RollingFeatures(stats=["ratio_min_max", "median"], window_sizes=4)
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, window_features=rolling, binner_kwargs={"n_bins": 3}
    )
    forecaster.fit(series=series, store_in_sample_residuals=False)
    results_residuals = forecaster.in_sample_residuals_
    results_residuals_bin = forecaster.in_sample_residuals_by_bin_
    results_binner_intervals = forecaster.binner_intervals_

    if encoding is not None:
        expected_residuals = {'1': None, '2': None, '_unknown_level': None}
        expected_residuals_by_bin = {'1': None, '2': None, '_unknown_level': None}
        expected_binner_intervals = {
            "1": {
                0: (7.268796921364613, 8.987998904684593),
                1: (8.987998904684593, 9.830020290458247),
                2: (9.830020290458247, 12.366369800747009),
            },
            "2": {
                0: (7.211558912361964, 9.22351564748152),
                1: (9.22351564748152, 11.350695147373349),
                2: (11.350695147373349, 11.844709741478395),
            },
            "_unknown_level": {
                0: (7.211558912361964, 9.016289033159751),
                1: (9.016289033159751, 10.730519759760567),
                2: (10.730519759760567, 12.366369800747009),
            },
        }

    else:
        expected_residuals = {'_unknown_level': None}
        expected_residuals_by_bin = {'_unknown_level': None}
        expected_binner_intervals = {
            "_unknown_level": {
                0: (6.81635531697439, 9.303028105789933),
                1: (9.303028105789933, 10.690051434777626),
                2: (10.690051434777626, 12.789680311772173),
            }
        }

    # In-sample residuals
    assert isinstance(results_residuals, dict)
    assert results_residuals.keys() == expected_residuals.keys()
    assert np.all(results_residuals[k] == expected_residuals[k] for k in results_residuals.keys())

    # In-sample residuals by bin
    assert isinstance(results_residuals_bin, dict)
    assert results_residuals_bin.keys() == expected_residuals_by_bin.keys()
    assert np.all(results_residuals_bin[k] == expected_residuals_by_bin[k] for k in results_residuals_bin.keys())

    # Binner intervals
    assert results_binner_intervals.keys() == expected_binner_intervals.keys()
    for level in results_binner_intervals.keys():
        assert results_binner_intervals[level].keys() == expected_binner_intervals[level].keys()
        for k in results_binner_intervals[level].keys():
            assert results_binner_intervals[level][k][0] == approx(expected_binner_intervals[level][k][0])
            assert results_binner_intervals[level][k][1] == approx(expected_binner_intervals[level][k][1])


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'ordinal_category', 'onehot', None], 
                         ids = lambda encoding: f'encoding: {encoding}')
def test_fit_in_sample_residuals_not_stored_probabilistic_mode_False(encoding):
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False` and _probabilistic_mode=False.
    """
    rng = np.random.default_rng(1894)
    series = {
        "1": pd.Series(rng.normal(10, 5, 20)), 
        "2": pd.Series(rng.normal(10, 5, 20))
    }

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding
    )
    forecaster._probabilistic_mode = False
    forecaster.fit(series=series, store_in_sample_residuals=False)
    results_residuals = forecaster.in_sample_residuals_
    results_residuals_bin = forecaster.in_sample_residuals_by_bin_
    results_binner_intervals = forecaster.binner_intervals_

    if encoding is not None:
        expected_residuals = {'1': None, '2': None, '_unknown_level': None}
        expected_residuals_by_bin = {'1': None, '2': None, '_unknown_level': None}
        expected_binner_intervals = {}

    else:
        expected_residuals = {'_unknown_level': None}
        expected_residuals_by_bin = {'_unknown_level': None}
        expected_binner_intervals = {}

    # In-sample residuals
    assert isinstance(results_residuals, dict)
    assert results_residuals.keys() == expected_residuals.keys()
    assert np.all(results_residuals[k] == expected_residuals[k] for k in results_residuals.keys())

    # In-sample residuals by bin
    assert isinstance(results_residuals_bin, dict)
    assert results_residuals_bin.keys() == expected_residuals_by_bin.keys()
    assert np.all(results_residuals_bin[k] == expected_residuals_by_bin[k] for k in results_residuals_bin.keys())

    # Binner intervals
    assert results_binner_intervals == expected_binner_intervals


def test_fit_last_window_stored():
    """
    Test that values of last window are stored after fitting.
    """
    series = {
        '1': pd.Series(np.arange(5, dtype=float)), 
        '2': pd.Series(np.arange(5, dtype=float))
    }

    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series)

    expected = {
        '1': pd.Series(
                 data  = np.array([2., 3., 4.]),
                 index = pd.RangeIndex(start=2, stop=5, step=1),
                 name  = '1',
                 dtype = float
             ),
        '2': pd.Series(
                 data  = np.array([2., 3., 4.]),
                 index = pd.RangeIndex(start=2, stop=5, step=1),
                 name  = '2',
                 dtype = float
             )
    }

    for k in forecaster.last_window_.keys():
        pd.testing.assert_series_equal(forecaster.last_window_[k], expected[k])


@pytest.mark.parametrize("encoding, encoding_mapping_", 
                         [('ordinal'         , {'1': 0, '2': 1}), 
                          ('ordinal_category', {'1': 0, '2': 1}),
                          ('onehot'          , {'1': 0, '2': 1}),
                          (None              , {'1': 0, '2': 1})], 
                         ids = lambda dt: f'encoding, mapping: {dt}')
def test_fit_encoding_mapping(encoding, encoding_mapping_):
    """
    Test the encoding mapping of _create_train_X_y.
    """
    series = {
        '1': pd.Series(np.arange(7, dtype=float)), 
        '2': pd.Series(np.arange(7, dtype=float))
    }
    
    forecaster = ForecasterRecursiveMultiSeries(
                     LinearRegression(),
                     lags     = 3,
                     encoding = encoding,
                 )
    forecaster.fit(series=series, suppress_warnings=True)
    
    assert forecaster.encoding_mapping_ == encoding_mapping_



# Unit test set_params ForecasterRecursiveClassifier
# ==============================================================================
from sklearn.linear_model import LogisticRegression
from skforecast.recursive import ForecasterRecursiveClassifier


def test_set_params():
    """
    """
    forecaster = ForecasterRecursiveClassifier(
        LogisticRegression(penalty='l1'), lags=3
    )
    new_params = {'penalty': 'elasticnet'}
    forecaster.set_params(new_params)
    expected = {
        'C': 1.0,
        'class_weight': None,
        'dual': False,
        'fit_intercept': True,
        'intercept_scaling': 1,
        'l1_ratio': None,
        'max_iter': 100,
        'multi_class': 'deprecated',
        'n_jobs': None,
        'penalty': 'elasticnet',
        'random_state': None,
        'solver': 'lbfgs',
        'tol': 0.0001,
        'verbose': 0,
        'warm_start': False
    }

    results = forecaster.estimator.get_params()
    
    assert results == expected
